// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// wrapper to include and enhance OpenCV 2/3.x.x's GPU/CUDA module
// ===============================================================

#pragma once

// account for changed headers, namespaces and class/method/function names in the GPU/CUDA module between version 2 and 3
// developer note #1: Most of OpenCV's headers encapsulate CUDA's runtime API, so this header can also be included if  
//                    CUDA is not installed. Do not break this by including OpenCV's "common.hpp", "NCV.hpp" or others 
//                    which do depend in "cuda_runtime.h". For converting a "cv::gpu::Stream" to a "cudaStream_t", 
//                    include "hawaii/GPU/CUDA.h".
// developer note #2: Feel free to add more classes renamed from "*_GPU" to "*_CUDA" as necessary, but make sure their 
//                    interface has not changed as well.
#include <opencv2/core/version.hpp>
#if CV_MAJOR_VERSION == 3
	#include <opencv2/core.hpp>
	#include <opencv2/cuda.hpp>
	#include <opencv2/cudaarithm.hpp>
	#include <opencv2/cudabgsegm.hpp>
	#include <opencv2/cudacodec.hpp>
	#include <opencv2/cudafeatures2d.hpp>
	#include <opencv2/cudafilters.hpp>
	#include <opencv2/cudaimgproc.hpp>
	#include <opencv2/cudaoptflow.hpp>
	#include <opencv2/cudastereo.hpp>
	#include <opencv2/cudawarping.hpp>
	namespace cv {
		namespace gpu = cv::cuda ;
		namespace cuda {
			typedef cv::cuda::BFMatcher_CUDA BFMatcher_GPU ;
			typedef cv::cuda::FAST_CUDA      FAST_GPU      ;
			typedef cv::cuda::ORB_CUDA       ORB_GPU       ;
		}
	}
#elif CV_MAJOR_VERSION == 2
	#include <opencv2/core/core.hpp>
	#include <opencv2/gpu/gpu.hpp>
#else
	#error "unsupported version of OpenCV"
#endif

namespace hawaii {
namespace GPU {

// forward declarations
class MemoryPool ;
extern MemoryPool memoryPool ;

// unified functions for asynchronous copying between CPU and GPU
void copyAsync( const cv::gpu::CudaMem         srcCudaMem,
                      cv::gpu::GpuMat&         dstGpuMat,
                      cv::gpu::Stream&         stream,
                      hawaii::GPU::MemoryPool& pool       = hawaii::GPU::memoryPool ) ;
void copyAsync( const cv::gpu::GpuMat          srcGpuMat,
                      cv::gpu::CudaMem&        dstCudaMem,
                      cv::gpu::Stream&         stream,
                      hawaii::GPU::MemoryPool& pool       = hawaii::GPU::memoryPool ) ;
void copyAsync( const cv::gpu::GpuMat          srcGpuMat,
                      cv::gpu::GpuMat&         dstGpuMat,
                      cv::gpu::Stream&         stream,
                      hawaii::GPU::MemoryPool& pool       = hawaii::GPU::memoryPool ) ;

// unified functions for asynchronous batch assignment
void assignAsync(       cv::gpu::GpuMat  dstGpuMat,
                  const cv::Scalar       value,
                        cv::gpu::Stream& stream    ) ;
void assignAsync(       cv::gpu::GpuMat  dstGpuMat,
                  const cv::gpu::GpuMat  mask,
                  const cv::Scalar       value,
                        cv::gpu::Stream& stream    ) ;

// types of GPU memory, used i.a. by "hawaii::AutoMat" and "hawaii::GPU::MemoryPool"
enum MemoryType {
	memoryOwn,    // GPU's own memory (might cause unnecessary copies in case it actually does not have any)
	memoryMapped, // page-locked CPU memory being mapped into the GPU address space (not write-combined)
	memoryAuto    // auto-select type based on hardware and/or an existing instance
} ;

// create an instance of "cv::{gpu,cuda}::Stream" which does not implicitly synchronize with the default stream
cv::gpu::Stream streamNonBlockingFactory() ;

} } // namespaces "hawaii::GPU"
