// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base classes for work items being passed through flow components
// ================================================================

#pragma once

#include "hawaii/common/optimization.h"
#include "hawaii/common/cPlusPlus11.h"
#include "hawaii/common/error.h"
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/smart_ptr/shared_ptr.hpp>
#include <boost/foreach.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"
#include <stdint.h>
#include <vector>
#include <stack>

namespace hawaii {
namespace flow {

// base partitioning information: derive from this only if your work item type requires additional information on how 
//                                to split or merge it
class PartInfoBase {
	public:
	virtual ~PartInfoBase() {}
} ;

// base work item type: You may prefer to derive from "ItemTemplate" below if partitioning should be implemented.
class ItemBase {
	
	// avoid copying overhead by actually using these rather than instances of the item itself
	public:
	typedef boost::shared_ptr< ItemBase > Ptr    ;
	typedef std::vector< Ptr >            PtrVec ;
	
	// require setting a running serial number: Be sure to do this correctly - either initialize an output item with the 
	//                                          input's serial number or (when implementing a source processing step) 
	//                                          maintain a local counter.
	public:
	ItemBase( const uint64_t serialNumberArg,
	          const double   ratioArg        = 1.0 ) :
		ratio(        ratioArg        ),
		serialNumber( serialNumberArg ) {
	}
	
	// split this item into partitions sized proportionally to "ratios": overriding implementations only needed if 
	// partitioning is actually used
	public:
	virtual void split(       ItemBase::PtrVec&      partsBase,
	                    const std::vector< double >& ratios,
	                    const PartInfoBase&          partInfoBase = PartInfoBase() ) ;
	
	// merge the given partitions into "whole": see comment above
	// user note: You can call this on any instance of the right type, e.g. one of the input partitions.
	// developer note: "this" cannot be used instead of "wholeBase" because then its allocation would have had to take 
	//                 place where its actual type is not known. Therefore, it does not matter on which instance you 
	//                 call this method as long as its actual type is right.
	public:
	virtual void merge( const ItemBase::PtrVec& partsBase,
	                          ItemBase::Ptr&    wholeBase,
	                    const PartInfoBase&     partInfoBase = PartInfoBase() ) ;
	
	// this partition's percentage of the original item (in case of heterogeneous partitioning), total number of 
	// partitions (stack for nested instances of "hawaii::flow::Partitioning")
	public:
	const double ratio ;
	std::stack< size_t > numParts ;
	
	// use biggest possible value running serial number as indicator to end processing
	public:
	       const uint64_t serialNumber ;
	static const uint64_t serialNumberFinal = 0xFFFFFFFFFFFFFFFFULL ; 
	bool isFinal() const { return this->serialNumber == ItemBase::serialNumberFinal ; }
	
	// ensure correct d'tor
	public:
	virtual ~ItemBase() {}
	
} ; // class "ItemBase"

// signal to stop processing
extern const ItemBase::Ptr itemFinalPtr ;

// templated base work item type: allows implementing partitioning methods with actually used instead of base types; 
//                                derive using the curiously recurring template pattern, i.e. 
//                                "class MyItem : public ItemTemplate< MyItem, MyPartInfo > { ... } ;" 
template< typename ItemUser, typename PartInfo = PartInfoBase >
class ItemTemplate : public ItemBase {
	
	// see above
	public:
	typedef boost::shared_ptr< ItemUser > Ptr    ;
	typedef std::vector< Ptr >            PtrVec ;
	
	// require setting both a running serial number and a partitioning ratio: see "ItemBase::ItemBase()"
	public:
	ItemTemplate( const uint64_t serialNumberArg,
	              const double   ratioArg        ) :
		ItemBase( serialNumberArg, ratioArg ) {
	}
	
	// partitioning methods to be implemented with actually used types
	public:
	virtual void splitImpl(       PtrVec&                parts,
	                        const std::vector< double >& ratios,
	                        const PartInfo&              partInfo = PartInfo() ) = 0 ;
	virtual void mergeImpl( const PtrVec&               parts,
	                              Ptr&                  whole,
	                        const PartInfo&             partInfo = PartInfo() ) = 0 ;
	
	// partitioning methods with base types wrapping the above ones
	public:
	virtual void split(       ItemBase::PtrVec&      partsBase,
	                    const std::vector< double >& ratios,
	                    const PartInfoBase&          partInfoBase = PartInfoBase() ) override final ;
	virtual void merge( const ItemBase::PtrVec&     partsBase,
	                          ItemBase::Ptr&        wholeBase,
	                    const PartInfoBase&         partInfoBase = PartInfoBase() ) override final ;
	
} ; // class template "ItemTemplate"

// "ItemTemplate": templated base work item type
// ---------------------------------------------

// wrap the above functions to implement the internally-used ones with base type arguments
template< typename ItemUser, typename PartInfo >
inline void ItemTemplate< ItemUser, PartInfo >::split(       ItemBase::PtrVec&      partsBase,
                                                       const std::vector< double >& ratios,
                                                       const PartInfoBase&          partInfoBase ) {
	// regular processing
	if( HAWAII_LIKELY( !this->isFinal() ) ) {
		
		// convert partitioning information
		const PartInfo* const partInfoUser = dynamic_cast< const PartInfo* >( &partInfoBase ) ;
		HAWAII_ERROR_CONDITIONAL( !partInfoUser, "got partitioning information of wrong type" ) ;
		
		// split
		PtrVec partsUser ;
		partsUser.reserve( ratios.size() ) ;
		this->splitImpl( partsUser, ratios, *partInfoUser ) ;
		
		// convert result
		partsBase.clear() ;
		partsBase.reserve( partsUser.size() ) ;
		BOOST_FOREACH( const Ptr& partUser, partsUser ) {
			HAWAII_ERROR_CONDITIONAL( !partUser, "got non-existent partition" ) ;
			const ItemBase::Ptr partBase = boost::static_pointer_cast< ItemBase >( partUser ) ;
			partsBase.push_back( partBase ) ;
		}
		
	// distribute final items
	} else {
		partsBase = ItemBase::PtrVec( ratios.size(), itemFinalPtr ) ;
	}
}
template< typename ItemUser, typename PartInfo >
inline void ItemTemplate< ItemUser, PartInfo >::merge( const ItemBase::PtrVec& partsBase,
                                                             ItemBase::Ptr&    wholeBase,
                                                       const PartInfoBase&     partInfoBase ) {
	// regular processing
	bool isFinal = false ;
	BOOST_FOREACH( const ItemBase::Ptr& partBase, partsBase ) { isFinal |= partBase->isFinal() ; }
	if( HAWAII_LIKELY( !isFinal ) ) {
		
		// convert partitioning information
		const PartInfo* const partInfoUser = dynamic_cast< const PartInfo* >( &partInfoBase ) ;
		HAWAII_ERROR_CONDITIONAL( !partInfoUser, "got partitioning information of wrong type" ) ;
		
		// convert input
		PtrVec partsUser ;
		partsUser.reserve( partsBase.size() ) ;
		BOOST_FOREACH( const ItemBase::Ptr& partBase, partsBase ) {
			HAWAII_ERROR_CONDITIONAL( !partBase, "got non-existent partition" ) ;
			const Ptr partUser = boost::dynamic_pointer_cast< ItemUser >( partBase ) ;
			HAWAII_ERROR_CONDITIONAL( !partUser, "got partition of wrong type" ) ;
			partsUser.push_back( partUser ) ;
		}
		
		// merge
		Ptr wholeUser ;
		this->mergeImpl( partsUser, wholeUser, *partInfoUser ) ;
		
		// convert result
		HAWAII_ERROR_CONDITIONAL( !wholeUser, "got non-existent whole item" ) ;
		wholeBase = boost::static_pointer_cast< ItemBase >( wholeUser ) ;
		
	// return final item
	} else {
		wholeBase = itemFinalPtr ;
	}
}

} } // namespaces "hawaii::flow"
