// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// distribution of a single work item to multiple processing units
// ===============================================================

#pragma once

#include "hawaii/GPU/nvccProhibit.h"
#include "hawaii/flow/flow.h"
#include <boost/thread/mutex.hpp>
#include <boost/serialization/shared_ptr.hpp>

namespace hawaii {
namespace flow {

// wrapper around identical functionality on different devices
class Partitioning : public FlowBase {
	
	// c'tor to directly add any number of flow components and other arguments for which "Partitioning::add()" is defined
	public:
	template< typename ... Args >
	Partitioning( Args&& ... args ) ;
	
	// add one flow component at a time
	public:
	template< typename ... Args >
	void add( hawaii::flow::FlowBase& flowNext,
	          Args&& ...              argsRest ) ;
	
	// set optionally given additional partitioning information
	protected:
	template< typename ... Args >
	void add( hawaii::flow::PartInfoBase& partInfoArg,
	          Args&& ...                  argsRest    ) ;
	template< typename ... Args >
	void add( hawaii::flow::PartInfoBase* partInfoArg,
	          Args&& ...                  argsRest    ) ;
	
	// allow the variadic template to set the class's name and to terminate
	protected:
	template< typename ... Args >
	void add( const std::string nameArg,
	                Args&& ...  argsRest ) ;
	void add() ;
	
	// reset internal state
	public:
	virtual void reset() ;
	
	// adapter for each wrapped flow component
	protected:
	class Adapter : public IfacePull, public IfacePush {
		
		// interfaces for corresponding flow component
		protected:
		virtual void pullImpl(          ItemBase::Ptr& itemPtr ) override ;
		virtual bool pullTryImpl(       ItemBase::Ptr& itemPtr ) override ;
		virtual void pushImpl(    const ItemBase::Ptr& itemPtr ) override ;
		virtual bool pushTryImpl( const ItemBase::Ptr& itemPtr ) override ;
		ItemBase::Ptr partPullPtr, partPushPtr ;
		double ratioPartPullPrev ;
		
		// Kalman filter to estimate "x" in "1 / throughput = ratio * x[0] + x[1]" for corresponding flow component
		protected:
		void perfModelPredict( double& throughputInverseFactor,
		                       double& throughputInverseOffset ) const ;
		void perfModelUpdate( const double throughputInverse,
		                      const double ratio,
		                      const double noise = 0.1       ) ;
		struct PerfModel{
			double x[2]    ; // state estimate
			double P[2][2] ; // state co-variance
			double Q[2][2] ; // system noise
			void reset() ;
		} perfModel ;
		hawaii::Timer timerThroughputInverse  ;
		bool          trackerUpdatedBeforePullTryFailed ;
		
		// allow access by class whose member this is (i.e. its "membee"), know the actual instance
		protected:
		friend class Partitioning ;
		Partitioning* membee ;
		Adapter( Partitioning* const membeeArg ) ;
	} ;
	
	// contain pointers to flow components instead of copies (accompanied by individual partitioning adapters), know preceding flow
	protected:
	FlowBase::PtrVec flows ;
	const FlowBase* flowPullPtr ;
	std::vector< Adapter > adapters ;
	bool adaptersSrcEmpty() const ;
	bool adaptersDstReady() const ;
	public:
	void setFlowPreceding( const FlowBase& flowPrecedingArg ) { this->flowPullPtr = &flowPrecedingArg ; }
	
	// connection between external interfaces and the above adapters, thread safety helpers
	protected:
	void pullToAdapters() ;
	void pushFromAdapters() ;
	boost::mutex              mutexSrc,     mutexDst     ;
	boost::condition_variable condSrcReady, condDstEmpty ;
	
	// partitioning: user-defined minimum partitioning ratio and additional information, run-time adaptation of ratios
	public:
	double ratioMin ;
	protected:
	boost::shared_ptr< PartInfoBase > partInfoPtr ;
	const std::vector< double > getRatios() ;
	
	// run this partitioner, wait for it to finish or directly stop it
	public:
	virtual void run()  override ;
	virtual void wait() override ;
	virtual void stop() override ;
	
	// It does not make sense for a partitioner to be a source and/or a sink.
	public:
	virtual bool isSource() const override { return false ; }
	virtual bool isSink()   const override { return false ; }
	
	// optionally limit throughput: You may get or set "INFINITY" for unlimited throughput.
	// developer note: This class enforces it internally rather than in "FlowBase" for being able to optimize power 
	//                 efficiency (see below). It is however less precise.
	protected:
	double throughputLimitPartitioning ;
	public:
	virtual double getThroughputLimit() const { return this->throughputLimitPartitioning ; }
	virtual void   setThroughputLimit( const double throughputLimit ) { this->throughputLimitPartitioning = throughputLimit ; }
	
	// time interval between performance model resets, and therefore re-activations of inefficient processors
	public:
	double timeReset ;
	protected:
	hawaii::Timer timerReset ;
	
	// estimate power consumption
	public:
	virtual double power() const override ;
	
	// print evaluation
	public:
	virtual void eval( size_t indentationName = 0,
	                   size_t indentationData = 0 ) const override ;
	
	// safely destruct mutexes and conditions
	public:
	virtual ~Partitioning() ;
	
} ; // class "Partitioning"

// "Partitioning": wrapper around identical functionality on different devices
// ---------------------------------------------------------------------------

// c'tor to directly add any number of flow components and other arguments for which "Partitioning::add()" is defined
template< typename ... Args >
inline Partitioning::Partitioning( Args&& ... args ) :
	FlowBase( "partitioning (unnamed)" ),
	flowPullPtr( nullptr ),
	ratioMin( 0.05 ),
	partInfoPtr( new PartInfoBase() ),
	throughputLimitPartitioning( INFINITY ),
	timeReset( 15000.0 ),
	timerReset( false ) {
	this->add( std::forward< Args >( args ) ... ) ;
}

// add one flow component at a time
template< typename ... Args >
inline void Partitioning::add( hawaii::flow::FlowBase& flowNext,
                               Args&& ...              argsRest ) {
	
	// prohibit sources or sinks
	HAWAII_ERROR_CONDITIONAL( flowNext.isSource() || flowNext.isSink(),
	                          "Sources or sinks don't make sense for partitioning." ) ;
	
	// create individual adapters, connection to flow component delayed until "add( void )" is called
	this->adapters.push_back( Adapter( this ) ) ;
	
	// wrap component and continue with next
	this->flows.push_back( FlowBase::Ptr( &flowNext, boost::serialization::null_deleter() ) ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

// set optionally given additional partitioning information
template< typename ... Args >
inline void Partitioning::add( hawaii::flow::PartInfoBase& partInfoArg,
                               Args&& ...                  argsRest    ) {
	this->partInfoPtr.reset( &partInfoArg, boost::serialization::null_deleter() ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}
template< typename ... Args >
inline void Partitioning::add( hawaii::flow::PartInfoBase* partInfoArg,
                               Args&& ...                  argsRest    ) {
	this->partInfoPtr.reset( partInfoArg ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

// allow the variadic template to set the class's name
template< typename ... Args >
inline void Partitioning::add( const std::string nameArg,
                                     Args&& ...  argsRest ) {
	this->name = nameArg ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

} } // namespaces "hawaii::flow"
