// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// classical pipeline
// ==================

#pragma once

#include "hawaii/GPU/nvccProhibit.h"
#include "hawaii/flow/flow.h"
#include "hawaii/flow/queue.h"
#include <boost/serialization/shared_ptr.hpp>

namespace hawaii {
namespace flow {

// wrapper around consecutive flow components
class Pipeline : public FlowBase {
	
	// c'tor to directly any number of flow components (with queues in between) and other arguments for which 
	// "Pipeline::add()" is defined
	public:
	template< typename ... Args >
	Pipeline( Args&& ... args ) ;
	
	// add an initial flow component or one pair of queue and flow component at a time
	public:
	template< typename ... Args >
	void add( hawaii::flow::FlowBase& flowFirst,
	          Args&& ...              argsRest  ) ;
	template< typename ... Args >
	void add( hawaii::flow::QueueBase& queueNext,
	          hawaii::flow::FlowBase&  flowNext,
	          Args&& ...               argsRest  ) ;
	template< typename ... Args >
	void add( hawaii::flow::QueueBase* queueNext,
	          hawaii::flow::FlowBase&  flowNext,
	          Args&& ...               argsRest  ) ;
	
	// allow the variadic template to set the class's name and to terminate
	protected:
	template< typename ... Args >
	void add( const std::string nameArg,
	                Args&& ...  argsRest ) ;
	void add() ;
	
	// reset internal state
	public:
	virtual void reset() ;
	
	// contain pointers to flow components and queues instead of copies
	protected:
	FlowBase::PtrVec  flows  ;
	QueueBase::PtrVec queues ;
	
	// run this pipeline, wait for it to finish or directly stop it
	public:
	virtual void run()  override ;
	virtual void wait() override ;
	virtual void stop() override ;
	virtual ~Pipeline() { this->wait() ; }
	
	// find whether this pipeline is a source and/or a sink
	virtual bool isSource() const override { return this->flows.front()->isSource() ; }
	virtual bool isSink()   const override { return this->flows.back()->isSink()    ; }
	
	// estimate power consumption
	public:
	virtual double power() const override ;
	
	// print evaluation
	public:
	virtual void eval( size_t indentationName = 0,
	                   size_t indentationData = 0 ) const override ;
	
} ; // class "Pipeline"

// "Pipeline": wrapper around consecutive flow components
// ------------------------------------------------------

// c'tor to directly any number of flow components (with queues in between) and other arguments for which 
// "Pipeline::add()" is defined
template< typename ... Args >
inline Pipeline::Pipeline( Args&& ... args ) :
	FlowBase( "pipeline (unnamed)" ) {
	this->add( std::forward< Args >( args ) ... ) ;
}

// add an initial flow component or one pair of queue and flow component at a time
template< typename ... Args >
inline void Pipeline::add( hawaii::flow::FlowBase& flowFirst,
                           Args&& ...              argsRest  ) {
	
	// check if initialization
	HAWAII_ERROR_CONDITIONAL( !this->flows.empty(),
	                          "Non-empty pipelines must be extended with pairs of 1st a queue and 2nd a flow component." ) ;
	
	// connect first flow component to adapters
	flowFirst.setSrcIface( *this ) ;
	flowFirst.setDstIface( *this ) ; // may be changed later
	
	// wrap component and continue with rest
	this->flows.push_back( FlowBase::Ptr( &flowFirst, boost::serialization::null_deleter() ) ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}
template< typename ... Args >
inline void Pipeline::add( hawaii::flow::QueueBase& queueNext,
                           hawaii::flow::FlowBase&  flowNext,
                           Args&& ...               argsRest  ) {
	// check if extension
	HAWAII_ERROR_CONDITIONAL( this->flows.empty(),
	                          "Empty pipelines must be initialized with a flow component, optionally followed by pairs of 1st a queue and 2nd a flow component." ) ;
	
	// connect last and next flow component via queue
	this->flows.back()->setDstIface( queueNext ) ;
	           flowNext.setSrcIface( queueNext ) ;
	           flowNext.setDstIface( *this     ) ;
	
	// wrap queue and component and continue with rest
	this->queues.push_back( QueueBase::Ptr( &queueNext, boost::serialization::null_deleter() ) ) ;
	this->flows.push_back(  FlowBase::Ptr(  &flowNext,  boost::serialization::null_deleter() ) ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}
template< typename ... Args >
inline void Pipeline::add( hawaii::flow::QueueBase* queueNext,
                           hawaii::flow::FlowBase&  flowNext,
                           Args&& ...               argsRest  ) {
	// check if extension
	HAWAII_ERROR_CONDITIONAL( this->flows.empty(),
	                          "Empty pipelines must be initialized with a flow component, optionally followed by pairs of 1st a queue and 2nd a flow component." ) ;
	
	// connect last and next flow component via queue
	this->flows.back()->setDstIface( *queueNext ) ;
	           flowNext.setSrcIface( *queueNext ) ;
	           flowNext.setDstIface( *this      ) ;
	
	// wrap queue and component and continue with rest
	this->queues.push_back( QueueBase::Ptr( queueNext ) ) ;
	this->flows.push_back( FlowBase::Ptr( &flowNext, boost::serialization::null_deleter() ) ) ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

// allow the variadic template to set the class's name
template< typename ... Args >
inline void Pipeline::add( const std::string nameArg,
                                 Args&& ...  argsRest ) {
	this->name = nameArg ;
	this->add( std::forward< Args >( argsRest ) ... ) ;
}

} } // namespaces "hawaii::flow"
