// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base classes for processing steps
// =================================

#pragma once

#include "hawaii/flow/flow.h"
#include "hawaii/common/helpers.h"
#include "hawaii/common/error.h"
#include "hawaii/GPU/nvccFixHeaderAbove.h"
#include <boost/smart_ptr/shared_ptr.hpp>
#include <boost/thread/thread.hpp>
#include "hawaii/GPU/nvccFixHeaderBelow.h"

namespace hawaii {
namespace flow {

// (really) base processing step: see "StepTemplate" below before deriving from this
class StepBase : public FlowBase {
	
	// pointer for classes wrapping processing steps
	public:
	typedef boost::shared_ptr< hawaii::flow::StepBase > Ptr    ;
	typedef std::vector< Ptr >                          PtrVec ;
	
	// actual processing function: Be sure to use "boost::{dynamic,static,const}_pointer_cast()" instead of 
	// "{static,dynamic}_cast" when implementing this!
	public:
	virtual void process( const hawaii::flow::ItemBase::Ptr& srcBasePtr,
	                            hawaii::flow::ItemBase::Ptr& dstBasePtr ) = 0 ;
	
	// c'tors: set multiple GPUs (or none) e.g. via a "std::vector< int >" initializer list or a single GPU via an "int"
	// developer note: Be sure to both check and use the hardware specified here! 
	public:
	StepBase( const std::string        nameArg     = "processing step (unnamed)",
	          const int                CPUCoresArg = 1,
	          const std::vector< int > GPUsArg     = std::vector< int >(),
	          const bool               sourceArg   = false,
	          const bool               sinkArg     = false                       ) ;
	StepBase( const std::string nameArg,
	          const int         CPUcoresArg,
	          const int         GPUArg,
	          const bool        sourceArg   = false,
	          const bool        sinkArg     = false ) ;
	
	// Is this processing step a source (i.e. produces work items instead of pulling them, e.g. by loading data from 
	// disk) and/or a sink (i.e. saves or displays results instead of passing them on)?
	protected:
	bool source, sink ;
	public:
	bool isSource() const { return this->source ; }
	bool isSink()   const { return this->sink   ; }
	
	// number of CPU cores to use
	protected:
	int CPUCores ;
	public:
	void setCPUCores( const int CPUCoresArg ) ;
	int getCPUCores() const {
		return this->CPUCores ;
	}
	
	// GPU(s) to use: GPU #n is used if the #n-th significant bit is set.
	protected:
	unsigned long long GPUBits ;
	public:
	void setGPUs( const std::vector< int > GPUsArg ) ;
	void setGPU(  const int                GPUArg  ) ;
	void addGPU(  const int                GPUArg  ) ;
	void remGPU(  const int                GPUArg  ) ;
	unsigned long long getGPUBits() const { return this->GPUBits ; }
	std::vector< int > getGPUs()    const ;
	int                getGPU()     const ;
	bool activateGPUTry() const ; // returns false if not exactly one GPU selected
	void activateGPU()    const ; // causes error "
	
	// run this processing step, wait for it to finish or directly stop it
	public:
	virtual void run()  override ;
	virtual void wait() override ;
	virtual void stop() override ;
	virtual ~StepBase() { this->wait() ; }
	
	// estimate power consumption
	protected:
	hawaii::Tracker1DPT1 trackerPower ;
	public:
	virtual double power() const override final { return this->trackerPower ; }
	
	// wrap a thread to do the processing
	protected:
	boost::thread thread ;
	void threadFunc() ;
	
} ; // class "StepBase"

// base processing step to directly use your work item type(s) instead of casting
template< typename ItemSrc,
          typename ItemDst = ItemSrc >
class StepTemplate : public StepBase {
	
	// shorten code
	public:
	typedef ItemSrc Src ; typedef boost::shared_ptr< Src > SrcPtr ;
	typedef ItemDst Dst ; typedef boost::shared_ptr< Dst > DstPtr ;
	
	// require purely virtual processing function with actually used instead of base types
	public:
	virtual void processImpl( const SrcPtr& srcPtr,
	                                DstPtr& dstPtr ) = 0 ;
	virtual void process( const ItemBase::Ptr& srcBasePtr,
	                            ItemBase::Ptr& dstBasePtr ) override final ;
	// c'tors: see above
	public:
	StepTemplate( const std::string        nameArg     = "processing step (unnamed)",
	              const int                CPUCoresArg = 1,
	              const std::vector< int > GPUsArg     = std::vector< int >(),
	              const bool               sourceArg   = false,
	              const bool               sinkArg     = false                       ) :
		StepBase( nameArg,
		          CPUCoresArg, GPUsArg,
		          sourceArg, sinkArg   ) {
	}
	StepTemplate( const std::string nameArg,
	              const int         CPUcoresArg,
	              const int         GPUArg,
	              const bool        sourceArg   = false,
	              const bool        sinkArg     = false ) :
		StepBase( nameArg,
		          CPUcoresArg, GPUArg,
		          sourceArg, sinkArg  ) {
	}
	
} ; // class template "StepTemplate"

// "StepTemplate": base processing step to directly use your work item type(s) instead of casting
// ----------------------------------------------------------------------------------------------

// wrap the actual-types processing function
template< typename ItemSrc, typename ItemDst >
inline void StepTemplate< ItemSrc, ItemDst >::process( const ItemBase::Ptr& srcBasePtr,
                                                             ItemBase::Ptr& dstBasePtr ) {
	// try to cast to actual work items
	const SrcPtr srcPtr = boost::dynamic_pointer_cast< Src >( srcBasePtr ) ;
			DstPtr dstPtr = boost::dynamic_pointer_cast< Dst >( dstBasePtr ) ;
	
	// only allow a source step to run without input
	HAWAII_ERROR_CONDITIONAL( !this->source && !srcBasePtr, "obtained non-existent source work item"  ) ;
	HAWAII_ERROR_CONDITIONAL( !this->source && !srcPtr,     "obtained source work item of wrong type" ) ;
	
	// if exactly one GPU selected, activate it
	this->activateGPUTry() ;
	
	// do the processing, "dstPtr" may have to be allocated there so update base pointer afterwards
	this->processImpl( srcPtr, dstPtr ) ;
	dstBasePtr = boost::static_pointer_cast< hawaii::flow::ItemBase >( dstPtr ) ;
	HAWAII_ERROR_CONDITIONAL( !this->sink && !dstBasePtr, "produced non-existent destination work item" ) ;
}

} } // namespaces "hawaii::flow"
