// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// processing step to load images from disk
// ========================================

#include "hawaii/IO/load.h"
#include "hawaii/common/optimization.h"
#include "hawaii/common/error.h"
#include <opencv2/highgui/highgui.hpp>
#include <cstdio>

// helper(s) only used here
namespace {
	
	// load all images for one filename index
	void loadImageSet( const std::vector< std::pair< std::string, std::string > >& imageDescs,
	                   const uint64_t                                              indexCurr,
	                   const uint64_t                                              indexEnd,
	                         hawaii::flow::ItemNamedImages::Ptr&                   dstPtr     ) {
		
		// reallocate as final item if index out of range
		if( indexCurr >= indexEnd ) {
			dstPtr = hawaii::flow::ItemNamedImages::Ptr(
				         new hawaii::flow::ItemNamedImages(
					         hawaii::flow::ItemBase::serialNumberFinal
				         )
			         ) ;
			return ;
		}
		
		// go through all image descriptions
		for( const auto& imageDesc : imageDescs ) {
			
			// load image
			const cv::Mat image = hawaii::IO::loadIndexed( imageDesc.first, indexCurr ) ;
			
			// reallocate as final item on failure
			if( HAWAII_UNLIKELY( image.empty() ) ) {
				dstPtr = hawaii::flow::ItemNamedImages::Ptr(
					         new hawaii::flow::ItemNamedImages(
						         hawaii::flow::ItemBase::serialNumberFinal
					         )
				         ) ;
				break ;
			}
			dstPtr->images[ imageDesc.second ] = image ;
		}
		
	} // function "loadImageSet()"
	
} // anonymous namespace

namespace hawaii {
namespace IO {

// load image with filename generated from pattern and index
cv::Mat loadIndexed( const std::string pattern,
                     const uint64_t    index   ) {
	
	// generate file name from pattern and index
	constexpr size_t bufferSize = 4096 ;
	char filename[ bufferSize ] ;
	HAWAII_ERROR_CONDITIONAL( 0 > snprintf( filename, bufferSize, pattern.c_str(), index ),
	                          "\"snprintf\" failed, increase \"bufferSize\""               ) ;
	
	// try to load image without converting channels ("-1")
	return cv::imread( filename, -1 ) ;
}

// implement inherited purely-virtual method to load the images
void StepLoad::processImpl( const SrcPtr& srcPtr,
                                  DstPtr& dstPtr ) {
	// check state
	HAWAII_ERROR_CONDITIONAL( this->imageDescs.empty(), "empty image list" ) ;
	
	// copy serial number from source item
	dstPtr = DstPtr( new Dst( srcPtr->serialNumber ) ) ;
	
	// depending on operation mode...
	if( !this->bufferAndRepeat ) {
		
		// simply load one set of images
		loadImageSet( this->imageDescs, this->indexCurr++, this->indexEnd, dstPtr ) ;
		
	} else {
		
		// on first call...
		if( this->bufferedItems.empty() ) {
			
			// pre-load all available image sets to the buffer
			hawaii::cout << this->name << ": filling buffer..." ;
			while( true ) {
				DstPtr bufPtr( new Dst( 0 ) ) ;
				loadImageSet( this->imageDescs, this->indexCurr++, this->indexEnd, bufPtr ) ;
				if( HAWAII_UNLIKELY( bufPtr->isFinal() ) ) { break ; }
				this->bufferedItems.push_back( bufPtr ) ;
				hawaii::cout << "." ;
			}
			hawaii::cout << " done" << hawaii::endl ;
		}
		
		// select next buffered item and...
		const DstPtr& bufPtr = this->bufferedItems[ dstPtr->serialNumber % this->bufferedItems.size() ] ;
		for( const auto& imageDesc : this->imageDescs ) {
			
			#if 0
				// ...copy CPU content to prevent modification of original one and to eliminate the slightest cache effects
				bufPtr->images[ imageDesc.second ].readCPU().copyTo( dstPtr->images[ imageDesc.second ].writeCPU() ) ;
			#else
				// ...directly use CPU content because the above copying can actually limit high-end systems
				dstPtr->images[ imageDesc.second ].writeCPU() = bufPtr->images[ imageDesc.second ].readCPU() ;
			#endif
		}
		
	} // if-else "buffer and repeat?"
	
} // method "StepLoad::process()"

} } // namespaces "hawaii::IO"
