// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base component of a graph-like network of processing steps, pipelines etc.
// ==========================================================================

#include "hawaii/flow/flow.h"
#include <thread>
#include <chrono>

namespace hawaii {
namespace flow {

// "FlowBase": abstract base e.g. for "hawaii::flow::Pipeline" and others
// ----------------------------------------------------------------------

// connection between external interface and internal components
void FlowBase::pullImpl( ItemBase::Ptr& itemPtr ) {
	
	// create a base work item if this is a source
	if( this->isSource() ) { itemPtr = ItemBase::Ptr( new ItemBase( this->serialNumberSource++ ) ) ; }
	else { HAWAII_ERROR_CONDITIONAL( !this->srcIface, "source interface required, but not specified" ) ; }
	
	// measure each contained flow's / thread's maximum possible throughput if it weren't limited below or outside, i.e. 
	// only measure the first time it tries
	this->trackerTIUUpdated.first.lock() ;
	if( this->trackerTIUUpdated.second.find( boost::this_thread::get_id() ) == this->trackerTIUUpdated.second.end() ) {
		this->trackerTIUUpdated.second.insert( boost::this_thread::get_id() ) ;
		if( HAWAII_LIKELY( this->timerThroughputInverse.isRunning() ) ) {
			this->trackerThroughputInverseUnconstrained( this->timerThroughputInverse.toc() ) ;
		}
	}
	this->trackerTIUUpdated.first.unlock() ;
	
	// obey speed limit
	if( this->timePullDelta > 0 ) {
		
		// atomic operation to allow concurrent usage
		hawaii::Timer::Rep timePullNextOld,
		                   timePullNextNew,
		                   timeCurrent,
		                   timeDeltaSleep  ;
		do { timePullNextOld = this->timePullNext ;
		     timeCurrent = hawaii::Timer::timeSinceEpoch() ;
		     if( timeCurrent > timePullNextOld ) { timePullNextNew = timeCurrent     + this->timePullDelta ; timeDeltaSleep = 0                             ; }
		     else                                { timePullNextNew = timePullNextOld + this->timePullDelta ; timeDeltaSleep = timePullNextOld - timeCurrent ; }
		} while( timePullNextOld != __sync_val_compare_and_swap( &this->timePullNext, timePullNextOld, timePullNextNew ) ) ;
		
		// wait if necessary
		if( timeDeltaSleep > 0 ) { std::this_thread::sleep_for( std::chrono::duration< double, std::milli >( timeDeltaSleep * hawaii::Timer::factorRepToMilliseconds ) ) ; }
	}
	
	// pull no matter if this is a source, so that external performance measurements know we did
	if( !this->srcIface ) {                                            this->trackerWaitedOnPull( 0.0 ) ;   }
	else                  { if( this->srcIface->pullTry( itemPtr ) ) { this->trackerWaitedOnPull( 0.0 ) ; }
	                        else                                     { this->trackerWaitedOnPull( 1.0 ) ;
	                                                                   this->srcIface->pull( itemPtr ) ;  } }
	// update internal throughput measurement
	this->trackerRatio( itemPtr->ratio ) ;
	if( HAWAII_LIKELY( this->timerThroughputInverse.isRunning() ) ) { this->trackerThroughputInverse( this->timerThroughputInverse.tocTic() ) ; }
	else                                                            {                                 this->timerThroughputInverse.tic()      ; }
	this->trackerTIUUpdated.first.lock() ;
	this->trackerTIUUpdated.second.erase( boost::this_thread::get_id() ) ;
	this->trackerTIUUpdated.first.unlock() ;
	
	// create item-specific latency timer
	{
		boost::lock_guard< boost::mutex > lock( this->timersLatency.first ) ;
		this->timersLatency.second[ itemPtr->serialNumber ].tic() ;
	}
}
bool FlowBase::pullTryImpl( ItemBase::Ptr& itemPtr ) {
	
	// create a base work item if this is a source
	if( this->isSource() ) { itemPtr = ItemBase::Ptr( new ItemBase( this->serialNumberSource++ ) ) ; }
	else { HAWAII_ERROR_CONDITIONAL( !this->srcIface, "source interface required, but not specified" ) ; }
	
	// measure each contained flow's / thread's maximum possible throughput if it weren't limited below or outside, i.e. 
	// only measure the first time it tries
	this->trackerTIUUpdated.first.lock() ;
	if( this->trackerTIUUpdated.second.find( boost::this_thread::get_id() ) == this->trackerTIUUpdated.second.end() ) {
		this->trackerTIUUpdated.second.insert( boost::this_thread::get_id() ) ;
		if( HAWAII_LIKELY( this->timerThroughputInverse.isRunning() ) ) {
			this->trackerThroughputInverseUnconstrained( this->timerThroughputInverse.toc() ) ;
		}
	}
	this->trackerTIUUpdated.first.unlock() ;
	
	// obey speed limit
	if( this->timePullDelta > 0 ) {
		
		// atomic operation to allow concurrent usage
		hawaii::Timer::Rep timePullNextOld,
		                   timePullNextNew,
		                   timeCurrent,
		                   timeDeltaSleep  ;
		timePullNextOld = this->timePullNext ;
		timeCurrent = hawaii::Timer::timeSinceEpoch() ;
		if( timeCurrent > timePullNextOld ) { timePullNextNew = timeCurrent     + this->timePullDelta ; timeDeltaSleep = 0                             ; }
		else                                { timePullNextNew = timePullNextOld + this->timePullDelta ; timeDeltaSleep = timePullNextOld - timeCurrent ; }
		
		// skip rather than wait, try to update next time only once
		if( timeDeltaSleep > 0
		 || !__sync_bool_compare_and_swap( &this->timePullNext, timePullNextOld, timePullNextNew ) ) {
			 return false ;
		}
	}
	
	// try to pull no matter if this is a source, so that external performance measurements know we did, skip on failure
	if( !this->srcIface ) {                                            this->trackerWaitedOnPull( 0.0 ) ;   }
	else                  { if( this->srcIface->pullTry( itemPtr ) ) { this->trackerWaitedOnPull( 0.0 ) ; }
	                        else                                     { this->trackerWaitedOnPull( 1.0 ) ;
	                                                                   return false ;                     } }
	// update internal throughput measurement
	this->trackerRatio( itemPtr->ratio ) ;
	if( HAWAII_LIKELY( this->timerThroughputInverse.isRunning() ) ) { this->trackerThroughputInverse( this->timerThroughputInverse.tocTic() ) ; }
	else                                                            {                                 this->timerThroughputInverse.tic()      ; }
	this->trackerTIUUpdated.first.lock() ;
	this->trackerTIUUpdated.second.erase( boost::this_thread::get_id() ) ;
	this->trackerTIUUpdated.first.unlock() ;
	
	// create item-specific latency timer
	{
		boost::lock_guard< boost::mutex > lock( this->timersLatency.first ) ;
		this->timersLatency.second[ itemPtr->serialNumber ].tic() ;
	}
	
	// success if not returned above
	return true ;
}
void FlowBase::pushImpl( const ItemBase::Ptr& itemPtr ) {
	
	// evaluate item-specific latency timer
	if( HAWAII_LIKELY( !itemPtr->isFinal() ) ) {
		
		// may already have been done in a previous failed "pushTry()"
		boost::lock_guard< boost::mutex > lock( this->timersLatency.first ) ;
		auto timerItem = this->timersLatency.second.find( itemPtr->serialNumber ) ;
		if( timerItem != this->timersLatency.second.end() ) {
			
			// update latency tracker
			const double latency = timerItem->second.toc() ;
			this->trackerLatency( latency ) ;
			this->timersLatency.second.erase( timerItem ) ;
			
			// also use latency as best initial guess for throughput timer
			if( HAWAII_UNLIKELY( !this->trackerThroughputInverse.isInit() ) ) {
				this->trackerThroughputInverse( this->trackerLatency ) ;
			}
		}
	}
	
	// pass on output work item no matter if this is a sink, so that external performance measurements know we did
	HAWAII_ERROR_CONDITIONAL( !this->isSink() && !this->dstIface, "destination interface required, but not specified" ) ;
	if( !this->dstIface ) {                           this->trackerWaitedOnPush( 0.0 ) ;   }
	else { if( this->dstIface->pushTry( itemPtr ) ) { this->trackerWaitedOnPush( 0.0 ) ; }
	       else                                     { this->trackerWaitedOnPush( 1.0 ) ;
	                                                  this->dstIface->push( itemPtr ) ;    } }
	// count all except final item
	if( HAWAII_LIKELY( !itemPtr->isFinal() ) ) { __sync_fetch_and_add( &this->counterItems, 1 ) ; }
}
bool FlowBase::pushTryImpl( const ItemBase::Ptr& itemPtr ) {
	
	// evaluate item-specific latency timer
	if( HAWAII_LIKELY( !itemPtr->isFinal() ) ) {
		
		// may already have been done in a previous failed "pushTry()"
		boost::lock_guard< boost::mutex > lock( this->timersLatency.first ) ;
		auto timerItem = this->timersLatency.second.find( itemPtr->serialNumber ) ;
		if( timerItem != this->timersLatency.second.end() ) {
			
			// update latency tracker
			const double latency = timerItem->second.toc() ;
			this->trackerLatency( latency ) ;
			this->timersLatency.second.erase( timerItem ) ;
			
			// also use latency as best initial guess for throughput timer
			if( HAWAII_UNLIKELY( !this->trackerThroughputInverse.isInit() ) ) {
				this->trackerThroughputInverse( this->trackerLatency ) ;
			}
		}	
	}
	
	// try to pass on output work item no matter if this is a sink, so that external performance measurements know we did, skip on failure
	HAWAII_ERROR_CONDITIONAL( !this->isSink() && !this->dstIface, "destination interface required, but not specified" ) ;
	if( !this->dstIface ) {                           this->trackerWaitedOnPush( 0.0 ) ;   }
	else { if( this->dstIface->pushTry( itemPtr ) ) { this->trackerWaitedOnPush( 0.0 ) ; }
	       else                                     { this->trackerWaitedOnPush( 1.0 ) ;
	                                                  return false ;                       } }
	// count all except final item
	if( HAWAII_LIKELY( !itemPtr->isFinal() ) ) { __sync_fetch_and_add( &this->counterItems, 1 ) ; }
	
	// success if not returned above
	return true ;
}

// print evaluation
// developer note: apply changes to "StepBaseGPU::eval()" as well
void FlowBase::eval( size_t indentationName,
                     size_t indentationData ) const {
	// prepare
	indentationData = std::max( indentationData, this->name.size() ) ;
	const std::string idtName(      indentationName * 4,                 ' ' ) ;
	const std::string idtDataFirst( indentationData - this->name.size(), ' ' ) ;
	const std::string idtDataOther( indentationData + 2,                 ' ' ) ;
	
	// print
	std::ostringstream lineRatio ;
	if( this->ratio() != 1.0 ) {
		lineRatio << idtName                       << idtDataOther << "ratio                      = " << this->ratio() * 100.0           << " %\n" ;
	}
	hawaii::cout << idtName << this->name << ": " << idtDataFirst << "latency                    = " << this->latency()                 << " ms\n"
	             << idtName                       << idtDataOther << "throughput                 = " << this->throughput()              << " Hz\n"
	             << idtName                       << idtDataOther << "throughput (unconstrained) = " << this->throughputUnconstrained() << " Hz\n"
	             << idtName                       << idtDataOther << "power                      = " << this->power()                   << " W\n"
	             << idtName                       << idtDataOther << "items                      = " << this->items()                   << "\n"
	             << idtName                       << idtDataOther << "waited on pull             = " << this->waitedOnPull() * 100.0    << " %\n"
	             << idtName                       << idtDataOther << "waited on push             = " << this->waitedOnPush() * 100.0    << " %\n"
	             << lineRatio.str()
	             << hawaii::endl ;
}

// c'tor with default initialization (non-default copy c'tor required by "boost::mutex"), reset internal state
FlowBase::FlowBase( const std::string nameArg ) :
	srcIface( nullptr                   ),
	dstIface( &hawaii::flow::terminator ),
	serialNumberSource( 0 ),
	name( nameArg ),
	counterItems( 0 ),
	timerThroughputInverse( false ),
	trackerLatency(                        0.95 ),
	trackerThroughputInverse(              0.95 ),
	trackerThroughputInverseUnconstrained( 0.95 ),
	trackerWaitedOnPull(                   0.95 ),
	trackerWaitedOnPush(                   0.95 ),
	trackerRatio(                          0.95 ),
	timePullDelta( 0 ),
	timePullNext( hawaii::Timer::timeSinceEpoch() ) {
}
void FlowBase::reset() {
	this->serialNumberSource = 0 ;
	this->counterItems = 0 ;
	this->timersLatency.first.lock()   ;
	this->timersLatency.second.clear() ;
	this->timersLatency.first.unlock() ;
	this->timerThroughputInverse.reset() ;
	this->trackerLatency.reset()                        ;
	this->trackerThroughputInverse.reset()              ;
	this->trackerThroughputInverseUnconstrained.reset() ;
	this->trackerWaitedOnPull.reset()                   ;
	this->trackerWaitedOnPush.reset()                   ;
	this->trackerRatio.reset()                          ;
	this->timePullNext = hawaii::Timer::timeSinceEpoch() ;
}

// optionally limit throughput: You may get or set "INFINITY" for unlimited throughput.
double FlowBase::getThroughputLimit() const {
	return 1000.0 / hawaii::Timer::factorRepToMilliseconds / this->timePullDelta ;
}
void FlowBase::setThroughputLimit( const double throughputLimit ) {
	HAWAII_ERROR_CONDITIONAL( throughputLimit < 0.0, "Throughput limit must be positive." ) ;
	this->timePullDelta = 1000.0 / hawaii::Timer::factorRepToMilliseconds / throughputLimit ;
}

} } // namespaces "hawaii::flow"
