// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// item to contain named "hawaii::AutoMat" images
// ==============================================

#include "hawaii/flow/itemNamedImages.h"
#include "hawaii/common/partitionize.h"
#include "hawaii/common/error.h"

namespace hawaii {
namespace flow {

// static constants
const int      PartInfoImageOverlap::typeSame   = 1                ;
const cv::Size PartInfoImageOverlap::windowNone = cv::Size( 1, 1 ) ;

// split into horizontal image stripes
void ItemNamedImages::splitImpl(       PtrVec&                parts,
                                 const std::vector< double >& ratios,
                                 const PartInfoImageOverlap&  partInfo ) {
	// pre-allocate
	for( const auto preAlloc : partInfo.preAllocs ) {
		
		// get and check reference image
		const cv::Mat& ref = this->images[ preAlloc.nameRef ] ;
		HAWAII_ERROR_CONDITIONAL( ref.empty(),
		                          "reference image not initialized" ) ;
		HAWAII_ERROR_CONDITIONAL( ref.rows < preAlloc.win.height
		                       || ref.cols < preAlloc.win.width,
		                          "reference image too small for window" ) ;
		
		// allocate in CPU memory, i.e. where merging will take place
		this->images[ preAlloc.name ].writeCPU( cv::Size( ref.cols - preAlloc.win.width  + 1,
		                                                  ref.rows - preAlloc.win.height + 1 ),
		                                        ( preAlloc.type == PartInfoImageOverlap::typeSame ? ref.type() : preAlloc.type ) ) ;
	}
	
	// reference number of rows from either the specified image or the one with the least rows
	int rowsRef = std::numeric_limits< int >::max() ;
	if( !partInfo.nameRef.empty() ) {
		rowsRef = this->images[ partInfo.nameRef ].rows ;
		HAWAII_ERROR_CONDITIONAL( rowsRef == 0,
		                          "reference image not initialized" ) ;
	} else {
		for( const auto& image : this->images ) {
			if( rowsRef > image.second.rows ) { rowsRef = image.second.rows ; }
		}
	}
	
	// find and apply partitioning
	const size_t numParts = ratios.size() ;
	parts.clear() ;
	parts.reserve( numParts ) ;
	for( const auto& rowRange : hawaii::partitionize( rowsRef, ratios, partInfo.winHeight - 1 ) ) {
		
		// allocate partition: Effective ratio may be slightly different from input due to pixel rounding.
		const double ratioPart = (double)rowRange.size() / (double)( rowsRef + ( partInfo.winHeight - 1 ) * ( numParts - 1 ) ) ;
		auto part = Ptr( new ItemNamedImages( this->serialNumber, this->ratio * ratioPart ) ) ;
		part->rowsRef = rowsRef ;
		
		// extract current partition from all images, adjust overlap if number of rows is different from the reference
		for( const auto& image : this->images ) {
			const int overlapDelta = image.second.rows - rowsRef ;
			part->images[ image.first ] = image.second.rowRange( rowRange.start, rowRange.end + overlapDelta ) ;
		}
		parts.push_back( part ) ;
	}
	
} // method "splitImpl()"

// merge horizontal image stripes
void ItemNamedImages::mergeImpl( const PtrVec&               parts,
                                       Ptr&                  whole,
                                 const PartInfoImageOverlap& partInfo ) {
	
	// reference sum of rows across partitions from either the specified image or the one with the least rows
	const int rowSumRef = this->rowsRef + ( partInfo.winHeight - 1 ) * ( parts.size() - 1 ) ;
	
	// allocate whole item: "ratioWhole" is not necessarily 1.0, e.g. with nested partitioning. 
	double ratioWhole = 0.0 ;
	for( const auto& part : parts ) {
		ratioWhole += part->ratio ;
	}
	whole = Ptr( new ItemNamedImages( parts[ 0 ]->serialNumber, ratioWhole ) ) ;
	
	// merge one image at a time
	for( const auto& imagePart0 : parts[ 0 ]->images ) {
		
		// helper vars
		const std::string name     = imagePart0.first  ;
		const cv::Mat     matPart0 = imagePart0.second.readCPU() ;
		
		// copy all partitions to CPU memory, check if they already reside in one common memory block
		bool oneBlock = true ;
		for( const auto& part : parts ) {
			if( part->images[ name ].readCPU().isSubmatrix() == false
			 || part->images[ name ].readCPU().datastart     != matPart0.datastart ) {
				oneBlock = false ;
				break ;
			}
		}
		
		// expand first partition to whole common memory block and continue with the next image
		if( oneBlock ) {
			cv::Mat matWhole = matPart0 ;
			cv::Size  sizeWhole   ;
			cv::Point offsetWhole ;
			matWhole.locateROI( sizeWhole, offsetWhole ) ;
			matWhole.adjustROI( offsetWhole.y, sizeWhole.height - offsetWhole.y - matWhole.rows,
			                    offsetWhole.x, sizeWhole.width  - offsetWhole.x - matWhole.cols ) ;
			whole->images[ name ] = matWhole ;
			continue ;
		}
		
		// deduce current image's overlap and merged number of rows from row sum across its partitions
		int rowSum = 0 ;
		for( const auto& part : parts ) {
			rowSum += part->images[ name ].rows ;
		}
		const int overlap = ( partInfo.winHeight - 1 ) + ( rowSum - rowSumRef ) / (int)parts.size() ;
		const int rows    = rowSum - overlap * ( parts.size() - 1 ) ;
		cv::Mat& matWhole = whole->images[ name ].writeCPU( rows, matPart0.cols, matPart0.type() ) ;
		
		// copy all of first partition
		int rowStart = 0 ;
		matPart0.copyTo( matWhole.rowRange( rowStart, rowStart + matPart0.rows ) ) ;
		rowStart += matPart0.rows ;
		
		// copy remaining partitions without overlap at top
		for( size_t partIdx = 1 ; partIdx < parts.size() ; ++partIdx ) {
			cv::Mat matPartN = parts[ partIdx ]->images[ name ].readCPU() ;
			matPartN = matPartN.rowRange( overlap, matPartN.rows ) ;
			matPartN.copyTo( matWhole.rowRange( rowStart, rowStart + matPartN.rows ) ) ; 
			rowStart += matPartN.rows ;
		}
		
	} // for "all images"
	
} // method "mergeImpl()"

} } // namespaces "hawaii::flow"
