// Copyright (C) 2013 by FZI Forschungszentrum Informatik am Karlsruher Institut fuer Technologie
// 
// Author: Benjamin Ranft (benjamin.ranft@web.de)
// 
// This file is part of libHawaii.
// 
// libHawaii is free software: you can redistribute it and/or modify it under the terms of the GNU General Public 
// License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later 
// version.
// 
// libHawaii is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied 
// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License along with libHawaii. If not, see 
// <http://www.gnu.org/licenses/>.


// base classes for processing steps
// =================================

#include "hawaii/flow/step.h"
#include "hawaii/GPU/CUDA.h"
#include "hawaii/common/hardware.h"
#include <sys/time.h>
#include <sys/resource.h>

// helpers only used here
namespace {
	
	// estimate power consumption every "itemsPerUpdate" items
	constexpr uint64_t itemsPerPowerUpdate  = 4,
	                   itemsPerPowerUpdate1 = itemsPerPowerUpdate - 1 ; 
	
	// get the CPU time (user and system) in [cores * micro-seconds] spent by this thread
	// user note: This does not see OpenMP threads spawned between calls!
	// developer note #1: "boost::cpu_timer" is not available before version 1.48. As long as 1.46 is still widely used, 
	//                    e.g. via Ubuntu 12.04, we skip portability and do this manually.
	// developer note #2: A copy of this code exists in "src/flow/stepGPU.cpp" - keep them in sync!
	uint64_t getSpentTimeCPU() {
		rusage temp ;
		getrusage( RUSAGE_THREAD, &temp ) ;
		return temp.ru_utime.tv_sec * 1000000 + temp.ru_utime.tv_usec
		     + temp.ru_stime.tv_sec * 1000000 + temp.ru_stime.tv_usec ;
	}

} // anonymous namespace

namespace hawaii {
namespace flow {

// "StepBase": (really) base processing step
// -----------------------------------------

// c'tors: set multiple GPUs (or none) e.g. via a "std::vector< int >" initializer list or a single GPU via an "int"
StepBase::StepBase( const std::string        nameArg,
                    const int                CPUCoresArg,
                    const std::vector< int > GPUsArg,
                    const bool               sourceArg,
                    const bool               sinkArg     ) :
	FlowBase( nameArg ),
	source( sourceArg ),
	sink(   sinkArg   ),
	CPUCores( 1    ),
	GPUBits(  0ULL ),
	trackerPower( 1.0 ) {
	this->setCPUCores( CPUCoresArg ) ;
	this->setGPUs(     GPUsArg     ) ;
}
StepBase::StepBase( const std::string nameArg,
                    const int         CPUCoresArg,
                    const int         GPUArg,
                    const bool        sourceArg,
                    const bool        sinkArg     ) :
	FlowBase( nameArg   ),
	source( sourceArg ),
	sink(   sinkArg   ),
	CPUCores( 1    ),
	GPUBits(  0ULL ),
	trackerPower( 1.0 ) {
	this->setCPUCores( CPUCoresArg ) ;
	this->setGPU(      GPUArg      ) ;
}

// number of CPU cores to use
void StepBase::setCPUCores( const int CPUCoresArg ) {
	HAWAII_ERROR_CONDITIONAL( CPUCoresArg < 1,                  "The number of CPU cores must be at least 1."                                      ) ;
	HAWAII_ERROR_CONDITIONAL( CPUCoresArg > hawaii::CPUThreads, "The number of CPU cores must be at most your system's maximum number of threads." ) ;
	this->CPUCores = CPUCoresArg ;
}

// GPU(s) to use: set multiple at once
void StepBase::setGPUs( const std::vector< int > GPUsArg ) {
	this->GPUBits = 0ULL ;
	for( const auto& GPU : GPUsArg ) { this->addGPU( GPU ) ; }
}

// GPU(s) to use: set one at a time
void StepBase::setGPU( const int GPUArg ) { HAWAII_ERROR_CONDITIONAL( (uint)GPUArg >= (uint)hawaii::GPUs, "invalid GPU index" ) ; this->GPUBits  =  ( 1ULL << GPUArg ) ; }
void StepBase::addGPU( const int GPUArg ) { HAWAII_ERROR_CONDITIONAL( (uint)GPUArg >= (uint)hawaii::GPUs, "invalid GPU index" ) ; this->GPUBits |=  ( 1ULL << GPUArg ) ; }
void StepBase::remGPU( const int GPUArg ) { HAWAII_ERROR_CONDITIONAL( (uint)GPUArg >= (uint)hawaii::GPUs, "invalid GPU index" ) ; this->GPUBits &= ~( 1ULL << GPUArg ) ; }

// GPU(s) to use: get selected one(s), 
std::vector< int > StepBase::getGPUs() const {
	std::vector< int > ret ;
	for( int GPU = 0 ; GPU < hawaii::GPUs ; ++GPU ) {
		if( this->GPUBits & ( 1ULL << GPU ) ) { ret.push_back( GPU ) ; }
	}
	return ret ;
}
int StepBase::getGPU() const {
	HAWAII_ERROR_CONDITIONAL( __builtin_popcountll( this->GPUBits ) != 1,
	                          "more or less than one GPU selected - use \"getGPUs\" instead" ) ;
	return __builtin_ctzll( this->GPUBits ) ; // dev. note: "ctz" = count trailing zeros
}

// (try to) activate single selected GPU
bool StepBase::activateGPUTry() const {
	if( __builtin_popcountll( this->GPUBits ) != 1 ) { return false ; }
	#ifndef HAWAII_NO_CUDA
		HAWAII_ERROR_CHECK_CUDA( cudaSetDevice( __builtin_ctzll( this->GPUBits ) ) ) ;
		return true ;
	#else
		return false ;
	#endif
}
void StepBase::activateGPU() const {
	HAWAII_ERROR_CONDITIONAL( !this->activateGPUTry(),
	                          "more or less than one GPU selected - don't know which one to activate" ) ;
}

// run this processing step, wait for it to finish or directly stop it
// developer note: Even though it says "StepBase::threadFunc", "run()" will use a derived class's overriding 
//                 implementation if available.
void StepBase::run() {
	this->stop() ;
	this->reset() ;
	this->thread = boost::thread( &StepBase::threadFunc, this ) ;
}
void StepBase::wait() {
	if( this->thread.joinable() ) {
		this->thread.join() ;
	}
}
void StepBase::stop() {
	if( this->thread.joinable() ) {
		this->thread.interrupt() ; // schedules to throw an exception at next interruption point
		this->thread.join() ;      // since "interrupt()" is non-blocking, wait until really finished
	}
}

// method to run within this processing step
void StepBase::threadFunc() {
	
	// if exactly one GPU selected, activate it
	this->activateGPUTry() ;
	
	// user and system CPU time spent by this step's thread, counter for missed updates (The information from the 
	// system call behind "getSpentTimeCPU()" sometimes does not update fast enough.)
	// developer note: A copy of this code exists in "src/flow/stepGPU.cpp" - keep them in sync!
	uint64_t spentTimeCPUsPrev = getSpentTimeCPU() ;
	uint64_t powerUpdatesSkipped = 0 ;
	
	// GPU time spent while "process()" is called
	// developer note: A copy of this code exists in "src/flow/stepGPU.cpp" - keep them in sync!
	cudaEvent_t eventGPUStart, eventGPUEnd ;
	#ifndef HAWAII_NO_CUDA
		HAWAII_ERROR_CHECK_CUDA( cudaEventCreate( &eventGPUStart, cudaEventDefault ) ) ;
		HAWAII_ERROR_CHECK_CUDA( cudaEventCreate( &eventGPUEnd,   cudaEventDefault ) ) ;
		double GPUsPower = 0.0 ;
		for( const auto GPU : this->getGPUs() ) {
			GPUsPower += hawaii::GPUPower( GPU ) ;
		}
	#endif
	
	// for catching a "boost::thread_interrupted"
	try {
		
		// stop via interruption point instead of loop logic
		while( true ) {
			
			// start each iteration with empty work items
			ItemBase::Ptr srcPtr, dstPtr ;
			
			// block until input work item obtained
			// developer note: no "if( this->isSource() ) ..." to always measure performance
			this->pull( srcPtr ) ;
			
			// decide about doing power consumption update
			const bool doPowerUpdate = this->counterItems % itemsPerPowerUpdate == itemsPerPowerUpdate1 ;
			
			// do not try to process final item, which might only point to an "ItemBase"
			if( HAWAII_LIKELY( srcPtr->isFinal() ) ) {
				dstPtr = itemFinalPtr ;
			} else {
				
				// enqueue taking time when GPU(s) can(!) start working, assuming that they also will
				#ifndef HAWAII_NO_CUDA
					if( doPowerUpdate && ( GPUsPower != 0.0 ) ) { HAWAII_ERROR_CHECK_CUDA( cudaEventRecord( eventGPUStart, 0 ) ) ; }
				#else
					(void)eventGPUStart ;
				#endif
				
				// do the processing
				this->process( srcPtr, dstPtr ) ;
				
				// enqueue taking the time GPU(s) is/are done
				#ifndef HAWAII_NO_CUDA
					if( doPowerUpdate && ( GPUsPower != 0.0 ) ) { HAWAII_ERROR_CHECK_CUDA( cudaEventRecord( eventGPUEnd,   0 ) ) ; }
				#else
					(void)eventGPUEnd  ;
				#endif
			}
			
			// block until output work item passed on (and optional throughput limit obeyed)
			// developer note: no "if( this->isSink() ) ..." to always measure performance
			this->push( dstPtr ) ;
			
			// estimate power consumption, assuming linear relations with each processor's thermal design power
			if( doPowerUpdate ) {
				
				// compute user and system CPU time spent since last time
				const uint64_t spentTimeCPUsCurr = getSpentTimeCPU() ;
				const uint64_t spentTimeCPUsDelta = spentTimeCPUsCurr - spentTimeCPUsPrev ;
				if( spentTimeCPUsDelta > 0 ) {
					
					// this processing step's cycle time in milli-seconds
					const double msCycle           = 1000.0 / this->throughput() ;
					const double msSinceLastUpdate = msCycle * itemsPerPowerUpdate * ( 1 + powerUpdatesSkipped ) ;
					
					// compute CPU usage ratio: extrapolate to user-specified number of used cores, normalize depending on 
					//                          hyper-threading
					const double ratioCPUs = spentTimeCPUsDelta * 0.001 / msSinceLastUpdate
					                       * (double)this->CPUCores / ( this->CPUCores > hawaii::CPUCores ?
					                     			                      (double)hawaii::CPUThreads        :
					                     			                      (double)hawaii::CPUCores          ) ;
					double power = ratioCPUs * hawaii::CPUPower ;
					
					// compute GPU usage ratio
					#ifndef HAWAII_NO_CUDA
						if( GPUsPower != 0.0 ) {
							HAWAII_ERROR_CHECK_CUDA( cudaEventSynchronize( eventGPUEnd ) ) ;
							float msGPUs ; HAWAII_ERROR_CHECK_CUDA( cudaEventElapsedTime( &msGPUs, eventGPUStart, eventGPUEnd ) ) ;
							const double ratioGPUs = (double)msGPUs / msCycle ;
							power += ratioGPUs * GPUsPower ;
						}
					#endif
					
					// update tracker, prepare next measurement
					this->trackerPower( power ) ;
					spentTimeCPUsPrev = spentTimeCPUsCurr ;
					powerUpdatesSkipped = 0 ;
				}
				
				// workaround in case timers have not yet been updates by the OS
				else { ++powerUpdatesSkipped ; }
			}
			
			// end thread on final item
			if( HAWAII_UNLIKELY( dstPtr->isFinal() ) ) {
				break ;
			}
			
			// allow thread to be interrupted at least once per loop
			boost::this_thread::interruption_point() ;
			
		} // while "thread not interrupted?"
		
	// end in a joinable way
	} catch( const boost::thread_interrupted& ) { /* just continue */ }
	
	// GPU time spent while "process()" is called
	// developer note: A copy of this code exists in "src/flow/stepGPU.cpp" - keep them in sync!
	#ifndef HAWAII_NO_CUDA
		HAWAII_ERROR_CHECK_CUDA( cudaEventDestroy( eventGPUStart ) ) ;
		HAWAII_ERROR_CHECK_CUDA( cudaEventDestroy( eventGPUEnd   ) ) ;
	#endif
	
} // method "StepBase::threadFunc()"

} } // namespaces "hawaii::flow"
